# coding: utf-8

__name__        = "kvaser"
__version__     = "0.1.5"
__license__     = "Apache Software License"
__description__ = "Various utility functions."
__author__      = u"Klaus Kähler Holst"
__email__       = "klaus@holst.it"
__url__         = "https://github.com/kkholst/kvaser"
__copyright__   = u"Copyright 2019-2022, Klaus Kähler Holst"
__summary__     = "Utility functions"
__keywords__    = "tools, utilities"

__all__ = [
    "__name__", "__summary__", "__description__",
    "__url__", "__version__", "__author__",
    "__email__", "__license__", "__copyright__",
]
